/*************************************************************************
*
* Name    :     vsscanf.cpp
*
* Author  : Anuj Seth <anujseth@yahoo.com>
*
* Purpose : A Linux-like vsscanf implementation
*
*                   VC++ provides the vsprintf() function but for some strange
*           reason does not provide the complementary vsscanf() function
*           which is available on most Unix systems. Several C++
*           implementations exist to plug this hole but most of them do not
*           provide the identical API available on Linux. This is an
*           implementation which enhances the CStringEx::Scanf extension
*           written by Rainer Bauer to provide an identical API to that
*           which is available on Linux.
*--------------------------------------------------------------------------
* Copyright (c) 2002, Anuj Seth and Rainer Bauer.
*--------------------------------------------------------------------------
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions
* are met:
*
* 1. Redistributions of source code must retain the above copyright
*    notice, this list of conditions and the following disclaimer.
*
* 2. All advertising materials mentioning features or use of this
*    software must display the following acknowledgment:
*    "This product includes software developed by Anuj Seth and Rainer
*    Bauer"
*
* 3. Redistributions of any form whatsoever must retain the following
*    acknowledgment:
*    "This product includes software developed by Anuj Seth and Rainer
*    Bauer"
*
*-------------------------------------------------------------------------
* This code is based on Rainer Bauer's CStringEx::Scanf extension class
* Website: http://codeguru.earthweb.com/string/StringExScanf.html
* CStringEx::Scanf extension Copyright (c) Rainer Bauer, 2001
*-------------------------------------------------------------------------
*
* THIS SOFTWARE IS PROVIDED ``AS IS'' BY THE AUTHORS AND INDEPENDENT
* CONTRIBUTORS AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
* LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR
* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
*************************************************************************/

// Comment the following line if you are not using this in an MFC based
// project in VC++ or have disable precompiled headers
#include "stdafx.h"
#include "vsscanf.h"
#include "LogFile.h"

#ifdef YP_INCLUDE_SCANF

// Note: when increasing no. of possible arguments,
// the switch statement in functions Scanf() must be extened!
#define CSTRINGEX_SCANF_MAX_ARGS        10

// Define some macros for readability
//
// Does anyone know how these macros can be replaced by one single macro ???
#define ADD_ARGS_1  p[0]
#define ADD_ARGS_2  p[0], p[1]
#define ADD_ARGS_3  p[0], p[1], p[2]
#define ADD_ARGS_4  p[0], p[1], p[2], p[3]
#define ADD_ARGS_5  p[0], p[1], p[2], p[3], p[4]
#define ADD_ARGS_6  p[0], p[1], p[2], p[3], p[4], p[5]
#define ADD_ARGS_7  p[0], p[1], p[2], p[3], p[4], p[5], p[6]
#define ADD_ARGS_8  p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7]
#define ADD_ARGS_9  p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7], p[8]
#define ADD_ARGS_10 p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7], p[8], p[9]
#endif

extern LogFile logFile;

void CStringEx::ReplaceStr( const char *crefPattern, const char *crefReplaceBy )
{
	int iPatternLen = (int)strlen(crefPattern);
	if( !iPatternLen )
		return;
	int iPos =  find(crefPattern);
	int iStartPos = 0;
	bool bModified = false;
	CStringEx strResult;
	while( iPos != -1 )
	{
		//signal there were replacements made
		bModified = true;
		strResult += substr((size_t)iStartPos, iPos-iStartPos) + crefReplaceBy;		
		
		//search for the next occurence of pattern
		iStartPos = iPos + iPatternLen;
		iPos = find(crefPattern, iStartPos);
	}
	// add the remaining string
	strResult += substr((size_t)iStartPos);
	
	//copy resulting string to ourselves, if replacements took place
	if( bModified )
		*this = strResult;
}

int CStringEx::SearchRegEx(char *str, const char *regexpr, char **startPos, const char *format, size_t sz)
{
	regexp *rx = regcomp(const_cast<char *>(regexpr));

	if(rx == NULL)
		return -1; // Invalid Regular Expression

	if(*startPos != NULL)
		str = *startPos;

	int retVal = regexec(rx, str, startPos);

	if(retVal == 1)
	{
		CStringEx fmt(format);
		CStringEx suffix(format);
		suffix.ReplaceStr("%%","%");
		suffix.Delete(0, suffix.find("%s")+2);
		CStringEx siz;
		siz.Format("%%%ds", sz);
		//logFile.Write(LOG_ADVANCED, "fmt=%s, siz = %s", fmt.c_str(), siz.c_str());
		fmt.ReplaceStr("%s", siz);
		//logFile.Write(LOG_ADVANCED, "after replace: fmt=%s", fmt.c_str());
		char *buf = (char *)calloc(1, sz);
		sscanf(*startPos, fmt.c_str(), buf);
		if (buf[sz-1] != '\0')
		{
			free(buf);
			free(rx);
			logFile.Write(LOG_BASIC, "CStringEx::SearchRegEx(): overflow occurred (data exceeded %d bytes)",
				sz);
			return -2; // overflow occurred
		}
		//logFile.Write(LOG_ADVANCED, "after sscanf: buf=%s", buf);
		CStringEx newstr(buf);
		free(buf);
		//logFile.Write(LOG_ADVANCED, "suffix=%s, newstr=%s", suffix.c_str(), newstr.c_str());
		if (suffix.length() > 0)
		{
			int i = newstr.find(suffix);
			if (i != npos)
			{
				newstr.Delete(i, newstr.length()-i);
				//logFile.Write(LOG_ADVANCED, "after deletion of suffix: newstr=%s", newstr.c_str());
			}
		}
		*this = newstr;
	}

	free(rx); // Free resources

	return !retVal; // regular expression not found
}

#ifdef YP_INCLUDE_SCANF

int CStringEx::sscanfWrapper( int numArgs, LPCTSTR format, void **p )
{
        switch ( numArgs )
        {
        case  0: return 0;
        case  1: return STSCANF( GetBuffer(0), format, ADD_ARGS_1  );
        case  2: return STSCANF( GetBuffer(0), format, ADD_ARGS_2  );
        case  3: return STSCANF( GetBuffer(0), format, ADD_ARGS_3  );
        case  4: return STSCANF( GetBuffer(0), format, ADD_ARGS_4  );
        case  5: return STSCANF( GetBuffer(0), format, ADD_ARGS_5  );
        case  6: return STSCANF( GetBuffer(0), format, ADD_ARGS_6  );
        case  7: return STSCANF( GetBuffer(0), format, ADD_ARGS_7  );
        case  8: return STSCANF( GetBuffer(0), format, ADD_ARGS_8  );
        case  9: return STSCANF( GetBuffer(0), format, ADD_ARGS_9  );
        case 10: return STSCANF( GetBuffer(0), format, ADD_ARGS_10 );
        }

        // When extending max. no. of arguments [CSTRINGEX_SCANF_MAX_ARGS],
        // this function must be updated!
        ASSERT( !TRUE );
        return 0;
}

int CStringEx::Scanf(int numArgs, LPCTSTR format, void **p)
{
        int numScanned;

        // Call sscanf with correct no. of arguments
        numScanned = sscanfWrapper( numArgs, format, p );

        return numScanned;
}

int CStringEx::vsScanf(LPCTSTR format, va_list arglist)
{
        return vsscanf(this->c_str(), format, arglist);
}

int vsscanf( const char *str, LPCTSTR format, va_list arglist)
{
        int             numArgs;
        int             numScanned;
        void            *argPointer[CSTRINGEX_SCANF_MAX_ARGS];
        LPTSTR          currBuff;
        _TXCHAR         currChar;

        CStringEx string(str);

        numArgs = 0;
        currBuff = TCSCHR( format, _TXCHAR( '%' ) );

        if( currBuff == NULL )
        {
                // No valid format specifier!
                ASSERT( !TRUE );
                return 0;
        }

        do
        {
                // Move pointer to next character
                currBuff = TCSINC( currBuff );
                currChar = _TXCHAR( *currBuff );

                if ( currChar == TCSNULL )
                {
                        // End of string
                        //      -> processing will stop!
                }
                else if ( currChar == _TXCHAR( '*' ) )
                {
                        // "%*" suppresses argument assignment
                        //      -> do not get argument from stack!
                }
                else if ( currChar == _TXCHAR( '%' ) )
                {
                        // "%%" substitutes "%" character!
                        //      -> do not get argument from stack!
                        //      -> Increment to next character

                        currBuff = TCSINC( currBuff );
                }
                else
                {
                        if ( numArgs >= CSTRINGEX_SCANF_MAX_ARGS )
                        {
                                // This function can only handle
                                // <CSTRINGEX_SCANF_MAX_ARGS> (10) arguments!
                                ASSERT( !TRUE );
                                return 0;
                        }

                        argPointer[numArgs++] = va_arg( arglist, void * );
                }

                currBuff = TCSCHR( currBuff, _TXCHAR( '%' ) );

        } while ( currBuff != NULL );

        va_end( arglist );

        // Call sscanf with correct no. of arguments
        numScanned = string.Scanf(numArgs, format, argPointer);

        return 1;
}
#endif
